-- Puffin: A Tojaso FowlPlay Project
-- Widget to display value bars (e.g., health and mana) and timer bars (e.g., auras)

-- Custom attributes:
-- fgTexture - texture to use for the bar foreground (default: color swatch) [c]
-- fgColor -- color to use for the bar foreground (default: blue) [d]
-- fgAlpha - opacity of the bar foreground (default: 1) [c]
-- bgTexture - texture to use for the bar background (default: same as foreground) [c]
-- bgColor - color to use for the bar background (default: same as foreground) [d]
-- bgAlpha - opacity of the bar background (default: 0.6) [c]
-- direction - which way foreground bar moves as value decreases, false=right-to-left, true=left-to-right (default: false) [c]
-- sparkAlpha - opacity of spark displayed at moving end of the foreground bar (default: 0) [c]
-- value - if defined then is the fraction of bar's foreground to fill and timer values are ignored (default: 1) [d]
-- duration - total time that the timer bar will run (default: none) [d]
-- maxTime - maximum time actually displayed on timer bar (default: none) [d]
-- timeLeft - remaining time on timer bar (default: none) [d]
-- flashPeriod - if defined and greater than 0 then flash rectangle over that number of seconds [d]
-- borderTexture - if defined then show a border around the filled area of the rectangle [d]
-- borderColor - color to apply to border (default: black) [d]
-- borderWidth - width of border (default: 8) [d]
-- borderOffset - offset of border from texture (default: 2) [d]
-- borderOffsetX, borderOffsetY - position offset for border and panel (default: none) [c]

local Widget_Bar = {}
local EF = Puffin.ErrorFormat
local recycledBars = {}
local activeBars = {}
local barNumber = 0
local defaultBlack = { r = 0, g = 0, b = 0, a = 1 }
local defaultBlue = { r = 0, g = 0, b = 1, a = 1 }
local borderTable = { tile = false, tileSize = 0, insets = { left = 0, right = 0, top = 0, bottom = 0 }}

-- Create a new bar widget object
function Widget_Bar.Create()
	local n, bar = next(recycledBars) -- get any available recycled widget
	if n then
		recycledBars[n] = nil
	else
		bar = {}
		barNumber = barNumber + 1
		local t = {}; bar._display = t -- table of display objects preserved when widget is recycled
		local frameName = "Puffin_Bar" .. tostring(barNumber) -- generate unique name for bar's frame
		t.container = CreateFrame("Frame", frameName .. "Container", UIParent)
		t.container:SetFrameStrata("MEDIUM")
		t.bgTexture = t.container:CreateTexture(nil, "BACKGROUND")
		t.fgTexture = t.container:CreateTexture(nil, "BORDER")	
		t.border = CreateFrame("Frame", frameName .. "Backdrop", t.container)
		t.spark = t.container:CreateTexture(nil, "OVERLAY")
		t.spark:SetTexture([[Interface\CastingBar\UI-CastingBar-Spark]])
		t.spark:SetBlendMode("ADD"); t.spark:SetTexCoord(0, 1, 0, 1)
	end
	table.insert(activeBars, bar) -- keep track of widgets so can animate if necessary
	return bar
end

-- Hide all the display objects associated with a bar
local function HideBar(t)
	t.border:SetBackdrop(nil); t.border:Hide(); t.fgTexture:Hide(); t.bgTexture:Hide(); t.spark:Hide(); t.container:Hide()
end

-- Release a bar when no longer needed, recycle object after clearing attributes and hiding display objects associated with it
function Widget_Bar:Release()
	for k in pairs(self) do if k ~= "_display" then self[k] = nil end end
	HideBar(self._display)
	for k, v in pairs(activeBars) do if v == self then activeBars[k] = nil end end
	table.insert(recycledBars, self)
end

-- Set display position for the widget relative to another widget or the overall screen
function Widget_Bar:SetPosition(anchor, anchorWidget, relative, offsetX, offsetY, region)
	local t, w = self._display, UIParent
	if anchorWidget then w = region and anchorWidget._display[region] or anchorWidget._display.container end
	t.container:ClearAllPoints(); t.container:SetPoint(anchor, w, relative, offsetX, offsetY)
end

-- Set the parent widget (used for managing display heirarchy)
function Widget_Bar:SetParent(widget)
	local t, w = self._display, widget._display
	if w and w.container then w = w.container else w = UIParent end
	t.container:SetParent(w)
	if not self.depth then local d = w:GetFrameLevel() + 1; t.container:SetFrameLevel(d); t.border:SetFrameLevel(d + 1) end
end

-- Compute fill percentage for a timer bar
local function TimerFill(t)
	local fill, now = 1, GetTime()
	if not t.startTime then t.startTime = now; t.offsetTime = t.duration - t.timeLeft end
	local maxTime = t.maxTime or t.duration
	local remaining = t.duration - (now - t.startTime + t.offsetTime) -- computing remaining time in seconds
	if remaining < 0 then remaining = 0 end -- make sure no rounding funnies
	if remaining > t.duration then remaining = t.duration end -- and no inaccurate durations!
	t.timeLeft = remaining -- update saved value
	if remaining < maxTime then fill = remaining / maxTime end -- calculate fraction of time remaining
	return fill
end

-- Calculate alpha for flashing, period is how long the total flash time should last
local function FlashAlpha(maxAlpha, period)
	if not period or period == 0 then return 1 end
	local frac = GetTime() / period
	frac = frac - math.floor(frac) -- get fractional part of current period
	if frac >= 0.5 then frac = 1 - frac end -- now goes from 0 to 0.5 then back to 0
	return (maxAlpha / 2) + (maxAlpha * frac) -- drops to half of maxAlpha
end

-- Configure the bar's display objects based on current attribute values (or defaults)
function Widget_Bar:Configure()
	assert(self.width and self.width > 0, EF("Widget_Bar", "undefined width"))
	assert(self.height and self.height > 0, EF("Widget_Bar", "undefined height"))
	local t = self._display
	local w, h, fill = self.width, self.height, 1	
	local fgAnchor, bgAnchor, sparkAnchor
	if self.direction then -- direction bar moves in, nil/false is "right to left" and true is "left to right"
		fgAnchor = "TOPRIGHT"; bgAnchor = "TOPLEFT"; sparkAnchor = "BOTTOMLEFT"
	else
		fgAnchor = "TOPLEFT"; bgAnchor = "TOPRIGHT"; sparkAnchor = "BOTTOMRIGHT"
	end

	if self.bgAlpha ~= 0 then
		t.bgTexture:ClearAllPoints(); t.bgTexture:SetPoint(bgAnchor, t.container, bgAnchor, 0, 0)
		local tex, c = self.bgTexture or self.fgTexture, self.bgColor or self.fgColor
		if tex then t.bgTexture:SetTexture(tex) else t.bgTexture:SetTexture(1, 1, 1) end
		if not c then c = defaultBlue end; t.bgTexture:SetVertexColor(c.r, c.g, c.b, 1)
		t.bgTexture:SetSize(w, h); t.bgTexture:SetTexCoord(0, 1, 0, 1); t.bgTexture:SetAlpha(self.bgAlpha or 0.6)
	end
		 
	if self.fgAlpha ~= 0 then
		t.fgTexture:ClearAllPoints(); t.fgTexture:SetPoint(fgAnchor, t.container, fgAnchor, 0, 0)
		if self.fgTexture then t.fgTexture:SetTexture(self.fgTexture) else t.fgTexture:SetTexture(1, 1, 1) end
		if self.fgColor then local c = self.fgColor; t.fgTexture:SetVertexColor(c.r, c.g, c.b, 1) end
		t.spark:SetSize(20, h); t.spark:ClearAllPoints(); t.spark:SetAlpha(self.sparkAlpha or 0)
		t.spark:SetPoint("TOP", t.fgTexture, bgAnchor, 0, 0); t.spark:SetPoint("BOTTOM", t.fgTexture, sparkAnchor, 0, 0)
	end
	
	t.container:SetSize(w, h); t.container:SetScale(self.scale or 1); HideBar(t)
end

-- Update the bar's value and color
function Widget_Bar:Update()
	local t = self._display
	if self.hidden or (self.condition == false) then
		HideBar(t)
	else	
		if self.bgAlpha ~= 0 then
			local c = self.bgColor or self.fgColor
			if not c then c = defaultBlue end; t.bgTexture:SetVertexColor(c.r, c.g, c.b, 1)
			t.bgTexture:SetAlpha(self.bgAlpha or 0.6); t.bgTexture:Show()
		else t.bgTexture:Hide() end

		local w, h, fill = self.width, self.height, 1	
		if self.fgAlpha ~= 0 then
			if self.fgColor then local c = self.fgColor; t.fgTexture:SetVertexColor(c.r, c.g, c.b, 1) end
			if self.value then fill = self.value elseif self.duration and self.timeLeft then fill = TimerFill(self) end
			if (w * fill) >= 1 then
				t.fgTexture:SetSize(w * fill, h)
				if self.direction then t.fgTexture:SetTexCoord(fill, 0, fill, 1, 0, 0, 0, 1) else
					t.fgTexture:SetTexCoord(0, 0, 0, 1, fill, 0, fill, 1) end
				local alpha = self.fgAlpha or 1 -- flashing only adjusts foreground alpha
				if self.flashPeriod then alpha = FlashAlpha(alpha, self.flashPeriod) end -- adjust alpha if flashing
				t.fgTexture:SetAlpha(alpha); t.fgTexture:Show()
				if self.sparkAlpha and self.sparkAlpha > 0 and fill < 1 then t.spark:Show() else t.spark:Hide() end
			else t.fgTexture:Hide(); t.spark:Hide() end
		end
		
		local d = self.depth; if d then t.container:SetFrameLevel(d); t.border:SetFrameLevel(d + 1) end
		t.container:SetAlpha(self.alpha or 1); t.container:Show()
		
		if self.borderTexture then
			local offset, edgeSize, c = self.borderOffset or 2, self.borderWidth or 8, self.borderColor or defaultBlack
			if (edgeSize < 0.1) then edgeSize = 0.1 end
			borderTable.edgeFile = self.borderTexture; borderTable.edgeSize = edgeSize;
			t.border:SetBackdrop(borderTable); t.border:SetBackdropBorderColor(c.r, c.g, c.b, c.a)
			t.border:SetSize(w + offset, h + offset)
			t.border:SetPoint("CENTER", t.bgTexture, "CENTER", self.borderOffsetX or 0, self.borderOffsetY or 0); t.border:Show()
		else
			t.border:SetBackdrop(nil); t.border:Hide()
		end	
	end
end

-- Update the bar's animated display objects (i.e., flashing, foreground bar and spark, overlay and artwork bars)
function Widget_Bar:Animate()
	if not self.hidden and (self.condition ~= false) then
		local t = self._display
		if self.fgAlpha ~= 0 then
			local w, h, fill = self.width, self.height, 1
			if self.value then fill = self.value elseif self.duration and self.timeLeft then fill = TimerFill(self) end
			if (w * fill) >= 1 then
				t.fgTexture:SetSize(w * fill, h)
				if self.direction then t.fgTexture:SetTexCoord(fill, 0, fill, 1, 0, 0, 0, 1) else
					t.fgTexture:SetTexCoord(0, 0, 0, 1, fill, 0, fill, 1) end
				local alpha = self.fgAlpha or 1 -- flashing only adjusts foreground alpha
				if self.flashPeriod then alpha = FlashAlpha(alpha, self.flashPeriod) end -- adjust alpha if flashing
				t.fgTexture:SetAlpha(alpha); t.fgTexture:Show()
				if self.sparkAlpha and self.sparkAlpha > 0 and fill < 1 then t.spark:Show() else t.spark:Hide() end
			else t.fgTexture:Hide(); t.spark:Hide() end
		end
	end
end

-- Run animations for all active widgets
function Widget_Bar:OnRefresh() for _, w in pairs(activeBars) do Widget_Bar.Animate(w) end end

-- Register this widget as a component so it can be found by views
Puffin.RegisterComponent("Widget_Bar", Widget_Bar, "WIDGET", 0)
