-- Puffin: A Tojaso FowlPlay Project
-- Widget to display rectangles with optional borders and flashing

-- Custom attributes:
-- texture - path to rectangle's texture (default: color swatch) [c]
-- texcoord - table containing texture's texcoord settings (default: { 0, 1, 0, 1 }) [c]
-- color - color to apply to texture (default: none) [d]
-- valueWidth - fraction of rectangle's width to fill (default: 0) [d]
-- valueHeight - fraction of rectangle's height to fill (default: 0) [d]
-- valueOffsetX - fraction of rectangle's width to offset horizontal position (default: 0) [d]
-- valueOffsetY - fraction of rectangle's height to offset vertical position (default: 0) [d]
-- valueAlpha - alpha setting for the rectangles interior (default: 0) [d]
-- minimumWidth - minimum width in pixels to fill (default: 0.5) [d]
-- minimumHeight - minimum height in pixels to fill (default: 0.5) [d]
-- flashPeriod - if defined and greater than 0 then flash rectangle over that number of seconds [d]
-- borderTexture - if defined then show a border around the filled area of the rectangle [d]
-- borderColor - color to apply to border (default: black) [d]
-- borderWidth - width of border (default: 8) [d]
-- borderOffset - offset of border from texture (default: 2) [d]
-- borderOffsetX, borderOffsetY - position offset for border and panel (default: none) [c]

local Widget_Rectangle = {}
local EF = Puffin.ErrorFormat
local recycledRectangles = {}
local activeRectangles = {}
local rectangleNumber = 0
local defaultBlack = { r = 0, g = 0, b = 0, a = 1 }
local borderTable = { tile = false, tileSize = 0, insets = { left = 0, right = 0, top = 0, bottom = 0 }}

-- Create a new rectangle widget object
function Widget_Rectangle.Create()
	local n, rect = next(recycledRectangles) -- get any available recycled widget
	if n then
		recycledRectangles[n] = nil
	else
		rect = {}
		rectangleNumber = rectangleNumber + 1
		local t = {}; rect._display = t -- table of display objects preserved when widget is recycled
		local frameName = "Puffin_Rectangle" .. tostring(rectangleNumber) -- generate unique name for frame
		t.container = CreateFrame("Frame", frameName .. "Container", UIParent)
		t.container:SetFrameStrata("MEDIUM")
		t.texture = t.container:CreateTexture(nil, "ARTWORK")
		t.border = CreateFrame("Frame", frameName .. "Backdrop", t.container)
		t.border:SetPoint("CENTER", t.container, "CENTER", 0, 0)
	end
	table.insert(activeRectangles, rect) -- keep track of widgets so can animate if necessary
	return rect
end

-- Hide all the display objects associated with the widget
local function HideRectangle(t) t.border:SetBackdrop(nil); t.border:Hide(); t.texture:Hide(); t.container:Hide() end

-- Release a bar when no longer needed, recycle object after clearing attributes and hiding display objects associated with it
function Widget_Rectangle:Release()
	for k in pairs(self) do if k ~= "_display" then self[k] = nil end end
	HideRectangle(self._display)
	for k, v in pairs(activeRectangles) do if v == self then activeRectangles[k] = nil end end
	table.insert(recycledRectangles, self)
end

-- Set display position for the widget relative to another widget or the overall screen
function Widget_Rectangle:SetPosition(anchor, anchorWidget, relative, offsetX, offsetY, region)
	local t, w = self._display, UIParent
	if anchorWidget then w = region and anchorWidget._display[region] or anchorWidget._display.container end
	t.container:ClearAllPoints(); t.container:SetPoint(anchor, w, relative, offsetX, offsetY)
	t.texture:SetAllPoints(t.container)
end

-- Set the parent widget (used for managing display heirarchy)
function Widget_Rectangle:SetParent(widget)
	local t, w = self._display, widget._display
	if w and w.container then w = w.container else w = UIParent end
	t.container:SetParent(w)
	if not self.depth then local d = w:GetFrameLevel() + 1; t.container:SetFrameLevel(d); t.border:SetFrameLevel(d + 1) end
end

-- Calculate alpha for flashing, period is how long the total flash time should last
local function FlashAlpha(maxAlpha, period)
	if not period or period == 0 then return 1 end
	local frac = GetTime() / period
	frac = frac - math.floor(frac) -- get fractional part of current period
	if frac >= 0.5 then frac = 1 - frac end -- now goes from 0 to 0.5 then back to 0
	return (maxAlpha / 2) + (maxAlpha * frac) -- drops to half of maxAlpha
end

-- Configure all the widget's display objects based on current configuration attribute values (or defaults)
function Widget_Rectangle:Configure()
	assert(self.width and self.width > 0, EF("Widget_Rectangle", "undefined width"))
	assert(self.height and self.height > 0, EF("Widget_Rectangle", "undefined height"))
	local t = self._display
	local w, h = self.width, self.height
	if self.texture then t.texture:SetTexture(self.texture) else t.texture:SetTexture(1, 1, 1) end -- plain color block
	if self.texcoord then t.texture:SetTexCoord(unpack(self.texcoord)) else t.texture:SetTexCoord(0, 1, 0, 1) end
	t.container:SetScale(self.scale or 1); HideRectangle(t)
end

-- Update the widget's display objects based on current data attribute values (or defaults)
function Widget_Rectangle:Update()
	local t = self._display
	local vw, vh, c = (self.valueWidth or 0) * self.width, (self.valueHeight or 0) * self.height, self.color
	if self.hidden or (self.condition == false) or vw < (self.minimumWidth or 0.5) or vh < (self.minimumHeight or 0.5) then
		HideRectangle(t)
	else
		if not c then c = defaultBlack end; t.texture:SetVertexColor(c.r, c.g, c.b, 1)	
		local vx, vy = (self.valueOffsetX or 0) * self.width, (self.valueOffsetY or 0) * self.height
		
		local alpha = self.alpha or 1 -- flashing only adjusts foreground alpha
		if self.flashPeriod then alpha = FlashAlpha(alpha, self.flashPeriod) end -- adjust alpha if flashing
		local d = self.depth; if d then t.container:SetFrameLevel(d); t.border:SetFrameLevel(d + 1) end
		t.container:SetAlpha(alpha); t.container:SetSize(vw, vh); t.container:Show()
		
		t.texture:ClearAllPoints(); t.texture:SetPoint("BOTTOMLEFT", t.container, "BOTTOMLEFT", vx, vy)
		t.texture:SetAlpha(self.valueAlpha or 1); t.texture:SetSize(vw, vh); t.texture:Show()
		
		if self.borderTexture then
			local offset, edgeSize, c = self.borderOffset or 2, self.borderWidth or 8, self.borderColor or defaultBlack
			if (edgeSize < 0.1) then edgeSize = 0.1 end
			borderTable.edgeFile = self.borderTexture; borderTable.edgeSize = edgeSize;
			t.border:SetBackdrop(borderTable); t.border:SetBackdropBorderColor(c.r, c.g, c.b, c.a)
			t.border:ClearAllPoints(); t.border:SetPoint("CENTER", t.texture, "CENTER", self.borderOffsetX or 0, self.borderOffsetY or 0)
			t.border:SetSize(vw + offset, vh + offset); t.border:Show()
		else
			t.border:SetBackdrop(nil); t.border:Hide()
		end	
	end
end

-- Run animations for all active widgets
function Widget_Rectangle:OnRefresh()
	for _, w in pairs(activeRectangles) do
		local vw, vh = w.valueWidth or 1, w.valueHeight or 1
		if not w.hidden and w.flashPeriod and vw > 0 and vh > 0 then
			local alpha = FlashAlpha(w.alpha or 1, w.flashPeriod) -- adjust alpha if flashing
			w._display.container:SetAlpha(alpha)
		end
	end
end

-- Register this widget as a component so it can be found by views
Puffin.RegisterComponent("Widget_Rectangle", Widget_Rectangle, "WIDGET", 0)
